"""
File Transfer - SMB Shares

This module contains actions for uploading & downloading files to/from an SMB/CIFS file server. It facilitates file sharing on windows machines using the SMB1 and SMB2 protocols and with linux machines using the Samba server application.
"""
import os
from typing import Callable

import smb
import timeout_decorator

from smb.SMBConnection import SMBConnection
from ppa import ui

from . import exceptions
from .records import Server

WORKSPACE_PATH = "/workspace/share"


def download(server: Server, share_name: str, source: str, destination: str) -> None:
    """Download a shared file from an SMB/CIFS file share to the task workspace.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __server:__ a [Server](../inputs_outputs/#server) dictionary

    - __share_name:__ the name of the shared folder

    - __source:__ the path to download the source file from on the file share

    - __destination:__ the name of the file to be saved in the task workspace

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - file_transfer.smb_share.download:
        destination: a_file.txt
        server:
          address: 233.252.0.25
          username: username@example.domain
          password: Password
          name: mem-1
        share_name: smb_share
        source: sub_folder/share_file.txt
    ```
    """

    client = _connect_to_server(_check_server_keys(server))
    file_path = os.path.join(WORKSPACE_PATH, destination)

    try:
        with open(file_path, "wb") as file:
            client.retrieveFile(service_name=share_name, path=source, file_obj=file)
    except smb.smb_structs.OperationFailure as e:
        _log_and_raise(exceptions.ShareDriveError, e.message)


def upload(server: Server, share_name: str, source: str, destination: str) -> None:
    """Upload a file from the task workspace to a folder on an SMB/CIFS file share.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __server:__ a [Server](../inputs_outputs/#server) dictionary

    - __share_name:__ the name of the shared folder

    - __source:__ the name of the file to be uploaded from the task workspace

    - __destination:__ the path to upload the source file to on the file share

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
      - file_transfer.smb_share.upload:
          destination: sub_folder/uploaded_file.txt
          server:
            address: 233.252.0.25
            username: username@example.domain
            password: Password
            name: mem-1
          share_name: smb_share
          source: a_file.txt
    ```
    """

    client = _connect_to_server(_check_server_keys(server))
    file_path = os.path.join(WORKSPACE_PATH, source)

    try:
        with open(file_path, "rb") as file:
            client.storeFile(service_name=share_name, path=destination, file_obj=file)
    except smb.smb_structs.OperationFailure as e:
        _log_and_raise(exceptions.ShareDriveError, e.message)


@timeout_decorator.timeout(20)
def _connect_to_server(server: Server):
    if (port := server.get("port", 139)) not in (valid_ports := [139, 445]):
        _log_and_raise(
            exceptions.InvalidPort, f"Port {port} is invalid, must be either {' or '.join([str(p) for p in valid_ports])}"
        )

    client = SMBConnection(
        username=server["username"],
        password=server["password"],
        my_name="client",
        remote_name=server["name"],
        use_ntlm_v2=True,
        sign_options=1,
    )
    try:
        assert client.connect(server["address"], port=port, timeout=10)
    except timeout_decorator.timeout_decorator.TimeoutError:
        _log_and_raise(exceptions.TimeoutError, f"Connection attempt timed out, failed to connect to {server['address']}")
    except AssertionError:
        _log_and_raise(exceptions.AuthentiactionFailed, "Authentication failed using the supplied username & password.")
    except smb.base.NotConnectedError:
        _log_and_raise(exceptions.ShareDriveError, f"Can't find the share drive {server['name']}")
    except Exception as e:
        if "Name or service not known" or "Name does not resolve" in str(e):
            _log_and_raise(exceptions.ServerNotFound, f"Cannot find server at {server['address']}")
        raise e
    return client


def _log_and_raise(exception: Callable, msg: str) -> None:
    ui.output_error(msg)
    raise exception(msg)


def _check_server_keys(server: Server) -> Server:
    server_keys = {"address", "username", "password", "name"}

    if missing_server_keys := [k for k in server_keys if k not in server]:
        _log_and_raise(exceptions.MissingKey, f"The key(s) {', '.join(missing_server_keys)} are missing from the server input")

    return server
